package DateIncrement;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw( NormalizeDateIncrement IncrementDate SetToday NextDate $today $thisYear $thisMonth $thisDay @repeatNames );

# ---------------------------------------------------------------------------------------------
# 
# datehash objects:
# date:    starting date to increment
# repeat:  the repeat frequency, see table above
# week:    week number 1-5, optional
# day:     name of day, Sunday to Saturday, optional
# ---------------------------------------------------------------------------------------------
# use
use strict;
use warnings;
use Date::Calc qw(:all);

# ---------------------------------------------------------------------------------------------
# public variables
our ($today, $thisYear, $thisMonth, $thisDay, @repeatNames);

# private variables
my (%repeatIncrements, %repeatAlias, %week, @localtime, $nextDate);


# ---------------------------------------------------------------------------------------------
# initialize
%repeatIncrements = (
	'daily'       => [ 0, 0, 1 ],
	'bidaily'       => [ 0, 0, 2 ],
	'weekday'     => [ 0, 0, 1 ],
	'weekend'     => [ 0, 0, 1 ],
	'weekly'      => [ 0, 0, 7 ],
	'biweekly'    => [ 0, 0, 14 ],
	'lunar'       => [ 0, 0, 28 ],
	'semimonthly' => [ 0, 0, 15 ],
	'monthly'     => [ 0, 1, 0 ],
	'bimonthly'   => [ 0, 2, 0 ],
	'quarterly'   => [ 0, 3, 0 ],
	'semiannual'  => [ 0, 6, 0 ],
	'annual'      => [ 1, 0, 0 ],
	'yesterday'   => [ 0, 0, -1 ],
);
@repeatNames = ( 'daily', 'bidaily', 'weekday', 'weekend', 'weekly', 'biweekly', 'lunar', 'semimonthy', 'monthly', 'bimonthly', 'quarterly', 'semiannual', 'annual' );
%repeatAlias
	= ( 'd' => 'daily', 'bd' => 'bidaily', 'w' => 'weekly', 'bw' => 'biweekly', 'l' => 'lunar', 'sm' => 'semimonthly', 'm' => 'monthly', 'bm' => 'bimonthly', 'q' => 'quarterly', 'sa' => 'semiannual', 'a' => 'annual' );
%week = ( 'weekday' => [ 1, 1, 1, 1, 1, 0, 0 ], 'weekend' => [ 0, 0, 0, 0, 0, 1, 1 ] );
@localtime = localtime();
( $thisYear, $thisMonth, $thisDay ) = ( $localtime[5] + 1900, $localtime[4] + 1, $localtime[3] );
$today = sprintf "%.4d-%.2d-%.2d", $thisYear, $thisMonth, $thisDay;
$nextDate = { date=>'', repeate=>'daily' };

# ---------------------------------------------------------------------------------------------
sub SetToday {    # for testing
	my ($date) = @_;
	$date = $$date{date} if ref($date) eq 'HASH';
	$today = $date;
	( $thisYear, $thisMonth, $thisDay ) = split m"-", $today;
}

# ---------------------------------------------------------------------------------------------
sub DateFromYMD {
	my ( $y, $m, $d ) = @_;
	return sprintf "%.4d-%.2d-%.2d", $y, $m, $d;
}

# ---------------------------------------------------------------------------------------------
sub DateToYMD {
	my ( $date, $setDay ) = @_;
	$date = $$date{date} if ref($date) eq 'HASH';
	my @YMD = split m"[/-]", $date;
	$YMD[2] = $setDay if $setDay;
	return @YMD;
}

# ---------------------------------------------------------------------------------------------
sub DateToYM {
	my ($date) = @_;
	$date = $$date{date} if ref($date) eq 'HASH';
	return ( DateToYMD($date) )[ 0 .. 1 ];
}

# ---------------------------------------------------------------------------------------------
sub SetDate {
	my ( $datehash, $y, $m, $d ) = @_;
	my ($date, $dow) = ( \($$datehash{date}) , \($$datehash{dow}));
	if ($y) {
		$$date = DateFromYMD( $y, $m, $d );
	} elsif (!$$date) {
		$$date ||= '1999-12-31';
	} elsif ( $$date =~ m"^(\d\d?)/(\d\d?)/?(\d*)$" ) {
		my ( $m, $d, $y ) = ( $1, $2, $3 );
		$y ||= $thisYear;
		$y += 2000 if $y < 100;
		$$date = DateFromYMD( $y, $m, $d );
	}
	$$dow = Day_of_Week(DateToYMD($$date));
}

# ---------------------------------------------------------------------------------------------
sub SetIncrement {
	my ($datehash) = @_;
	my $repeat = \( $$datehash{repeat} );
	$$datehash{week} = $1 if $repeat =~ s"[wW](\d)"";
	$$datehash{day} = $1 if $repeat =~ s"d(mon|tue|wed|thu|fri|sat|sun)[a-zA-Z]""i;	
	
	$$repeat = lc $$repeat;
	if ( !$repeatIncrements{$$repeat} ) {
		if ( $repeatAlias{$$repeat} ) {
			$$repeat = $repeatAlias{$$repeat};
		} else {
			foreach my $alias ( sort keys %repeatAlias ) {
				$$repeat = $repeatAlias{$alias} if $$repeat =~ m"^$alias";
			}
		}
	}
}

# ---------------------------------------------------------------------------------------------
sub SetDOW {
	my ($datehash) = @_;
	return $$datehash{week} = '' if !$$datehash{day};
	$$datehash{setdow} = Decode_Day_of_Week( $$datehash{day} );
	$$datehash{day}    = Day_of_Week_to_Text( $$datehash{setdow} );
	$$datehash{week}   = 'Last' if $$datehash{week} and $$datehash{week} =~ m"L"i;
}

# ---------------------------------------------------------------------------------------------
sub NthWeekdayOfMonth {
	my ( $datehash, $week ) = @_;
	return Nth_Weekday_of_Month_Year( DateToYM( $$datehash{date} ), $$datehash{setdow}, $week || $$datehash{week} );
}

# ---------------------------------------------------------------------------------------------
sub SetIncrementedDate {
	my ( $datehash, $setDay, @increments ) = @_;
	if ( $setDay and $setDay !~ m"^\d+$" ) {
		unshift @increments, ($setDay);
		$setDay = '';
	}
	@increments = ( $$datehash{repeat} ) if !@increments;
	foreach my $increment (@increments) {
		my @increment = ( 0, 0, 1 );
		@increment = @{$increment} if ref($increment) eq 'ARRAY';
		@increment = @{ $repeatIncrements{$increment} } if $repeatIncrements{$increment};
		SetDate( $datehash, Add_Delta_YMD( DateToYMD( $datehash, $setDay ), @increment ) );
	}
}

# ---------------------------------------------------------------------------------------------
sub MoveToNthWeekday {
	my ($datehash) = @_;
	return SetDate($datehash,NthWeekdayOfMonth($datehash)) if $$datehash{week} !~ m"^L"i;
	my @last = NthWeekdayOfMonth( $datehash, 5 );
	@last = NthWeekdayOfMonth( $datehash, 4 ) if !@last;
	SetDate( $datehash, @last );
}

# ---------------------------------------------------------------------------------------------
sub MoveToDayOfWeek {
	my ($datehash) = @_;
	return if !$$datehash{day};
	return MoveToNthWeekday($datehash) if $$datehash{week};
	SetDate($datehash);
	return if $$datehash{dow} == $$datehash{setdow};
	SetIncrementedDate( $datehash, [ 0, 0, $$datehash{setdow} > $$datehash{dow} ? $$datehash{setdow} - $$datehash{dow} : 7 + $$datehash{setdow} - $$datehash{dow} ] );
}

# ---------------------------------------------------------------------------------------------
sub NormalizeDateIncrement {
	my ($datehash) = @_;
	$$datehash{normalized} = 1;
	foreach ( 'date', 'repeat', 'week', 'day', 'dow' ) { $$datehash{$_} ||= ''; $$datehash{old}{$_} = $$datehash{$_}; }
	my ( $date, $repeat, $week, $dayname ) = ( \( $$datehash{date} ), \( $$datehash{repeat} ), \( $$datehash{week} ), \( $$datehash{day} ) );
	SetDate($datehash);
	SetIncrement($datehash);
	SetDOW($datehash);
	MoveToDayOfWeek($datehash);
	SetIncrementedDate($datehash,[0,0,$$datehash{offset}]) if $$datehash{offset};
}

# ---------------------------------------------------------------------------------------------
sub IncrementDateHash {
	my ($datehash,$doingOffset) = @_;
	NormalizeDateIncrement($datehash) if !$$datehash{normalized};
	my $offset = $$datehash{offset};
	if ($offset and !$doingOffset) {
		SetIncrementedDate($datehash,[0,0,-$offset]);
		IncrementDateHash($datehash,1);
		SetIncrementedDate($datehash,[0,0,$offset]);
		return;		
	}
	my $date = \( $$datehash{date} );
	my $dow = \$$datehash{dow};
	$$datehash{previous}{date} = $$date;
	my $repeat = $$datehash{repeat};
	return SetIncrementedDate($datehash) if $repeat =~ m"^(daily|bidaily|weekly|biweekly|lunar)$";
	if ( $repeat =~ m"^(weekend|weekday)$" ) {
		do { SetIncrementedDate($datehash); } until ( $week{$repeat}[ $$dow ] );
		return;
	}
	my ( $y, $m, $d ) = DateToYMD($datehash);
	my $EOM = Days_in_Month( $y, $m );
	if ( $repeat eq "semimonthly" ) {
		return SetDate( $datehash, $y, $m, $d + 15 < $EOM ? $d + 15 : $EOM ) if $d < 15;
		return SetDate( $datehash, $y, $m, $EOM ) if $d == 15;
		return SetIncrementedDate( $datehash, 15, 'monthly' ) if $d == $EOM;
		return SetIncrementedDate( $datehash, $d - 15, 'monthly' );
	}
	if ($$datehash{week}) {
		SetIncrementedDate( $datehash, 1 );
		MoveToNthWeekday($datehash);
		return;
	}
	return SetIncrementedDate($datehash) if $d != $EOM or $$datehash{repeat} eq 'annual';
	return SetIncrementedDate( $datehash, 1, 'monthly', 'monthly', 'yesterday' );
}

# ---------------------------------------------------------------------------------------------
sub IncrementDate {
	my ( $date, $increment ) = @_;
	return IncrementDateHash($date) if ref($date);
	my $datehash = { date => $date, repeat => $increment };
	IncrementDateHash($datehash);
	return $$datehash{date};
}
	
# ---------------------------------------------------------------------------------------------
sub NextDate {
	my ($date) = @_; # yyyy-mm-dd
	$$nextDate{date} = $date;
	IncrementDateHash($nextDate);
	return $$nextDate{date};
}

# ---------------------------------------------------------------------------------------------
# End
1;
__END__
