package PDFUtil;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw(
	PORTRAIT LANDSCAPE
	PDFOpen
	PDFFontSize
	PDFHeaders
	PDFClose
	PDFPrintLine
	PDFPrint
	PDFPrintHeader
	PDFPrintSubHeader
	PDFColumnHeader
	PDFDrawLine
	PDFLinePosition
	PDFColumnPosition
);
use strict;
use warnings;
use File::Basename;
use PDF::Create;
use constant PORTRAIT  => 0;
use constant LANDSCAPE => 1;

# ---------------------------------------------------------------------------------------------
# PDF
sub PDFOpen {
	my ( $File, $Landscape, $Title, $ColumnHeader, @PageSubHeaders ) = @_;
	my %PDF;
	my $MarginInches = 0.25;
	$PDF{Filename}       = UniqueFilename($File,'pdf');
	$PDF{Title}          = $Title;
	$PDF{Landscape}      = $Landscape;
	$PDF{ColumnHeader}   = $ColumnHeader;
	$PDF{PageSubHeaders} = \@PageSubHeaders;
	$PDF{pdf}            = new PDF::Create(
		filename     => $PDF{Filename},
		Author       => 'Bruce Reed',
		Title        => $Title,
		CreationDate => [ localtime() ],
	);
	$PDF{Font} = $PDF{pdf}->font( BaseFont => 'Courier' );
	$PDF{BoldFont} = $PDF{pdf}->font( BaseFont => 'Courier-Bold' );
	$PDF{FontSize} = 12;
	my $PageSize = $PDF{pdf}->get_page_size('Letter');
	my $DPI = $$PageSize[2] / 8.5;
	( $$PageSize[2], $$PageSize[3] ) = ( $$PageSize[3], $$PageSize[2] ) if $Landscape;    # Landscape
	( undef, undef, $PDF{PaperWidth}, $PDF{PaperLength} ) = @$PageSize;
	$PDF{Margin}       = $DPI * $MarginInches;
	$PDF{PrintLength}  = $PDF{PaperLength} - ( $PDF{Margin} * 2 );
	$PDF{PrintWidth}   = $PDF{PaperWidth} - ( $PDF{Margin} * 2 );
	$PDF{LeftMargin}   = $PDF{Margin};
	$PDF{Center}       = $PDF{PaperWidth} / 2;
	$PDF{RightMargin}  = $PDF{PaperWidth} - $PDF{Margin};
	$PDF{LinesPerPage} = $PDF{PrintLength} / $PDF{FontSize};
	$PDF{RootPage}     = $PDF{pdf}->new_page( MediaBox => $PageSize );
	my ( undef, undef, undef, $Day, $Month, $Year ) = localtime();
	$PDF{Date}       = sprintf "%.2d/%.2d/%.4d", $Month + 1, $Day, $Year + 1900;
	$PDF{PageNumber} = 0;
	$PDF{LineNumber} = $PDF{LinesPerPage} + 1;
	return \%PDF;
}

sub PDFFontSize {
	my ($PDF,$FontSize) = @_;
	$$PDF{FontSize} = $FontSize;
	$$PDF{LinesPerPage} = $$PDF{PrintLength} / $$PDF{FontSize};
	$$PDF{LineNumber} = $$PDF{LinesPerPage} + 1;
}

sub PDFHeaders {
	my ( $PDF, $Title, $ColumnHeader, @PageSubHeaders ) = @_;
	$$PDF{Title}          = $Title;
	$$PDF{ColumnHeader}   = $ColumnHeader;
	$$PDF{PageSubHeaders} = \@PageSubHeaders;
	$$PDF{LineNumber}     = $$PDF{LinesPerPage} + 1;
}

sub PDFClose {
	my ($PDF) = @_;
	$$PDF{pdf}->close;
}

sub PDFPrintLine {
	my ( $PDF, $Content, $Align, $Bold ) = @_;
	$$PDF{LineNumber}++;
	PDFPrintHeader($PDF) if $$PDF{LineNumber} > $$PDF{LinesPerPage};
	$$PDF{LinePosition} = PDFLinePosition( $PDF, $$PDF{LineNumber} );
	PDFPrint( $PDF, $Content, $Align, '', $Bold ) if defined $Content;
}

sub PDFPrint {    # PDFPrintLine is called first to set new line
	my ( $PDF, $Content, $Align, $Column, $Bold ) = @_;
	$Align ||= 'L';
	$Align = uc substr( $Align, 0, 1 );
	my $Font = $Bold ? 'BoldFont' : 'Font';
	if ( !$Column ) {
		$$PDF{Page}->stringl( $$PDF{$Font}, $$PDF{FontSize}, $$PDF{LeftMargin}, $$PDF{LinePosition}, $Content ) if $Align eq 'L';
		$$PDF{Page}->stringc( $$PDF{$Font}, $$PDF{FontSize}, $$PDF{Center}, $$PDF{LinePosition}, $Content ) if $Align eq 'C';
		$$PDF{Page}->stringr( $$PDF{$Font}, $$PDF{FontSize}, $$PDF{RightMargin}, $$PDF{LinePosition}, $Content ) if $Align eq 'R';
		return;
	}
	my $ColumnPosition = PDFColumnPosition( $PDF, $Column );
	$$PDF{Page}->stringl( $$PDF{$Font}, $$PDF{FontSize}, $ColumnPosition, $$PDF{LinePosition}, $Content ) if $Align eq 'L';
	$$PDF{Page}->stringc( $$PDF{$Font}, $$PDF{FontSize}, $ColumnPosition, $$PDF{LinePosition}, $Content ) if $Align eq 'C';
	$$PDF{Page}->stringr( $$PDF{$Font}, $$PDF{FontSize}, $ColumnPosition, $$PDF{LinePosition}, $Content ) if $Align eq 'R';
}

sub PDFPrintHeader {
	my ($PDF) = @_;
	$$PDF{PageNumber}++;
	$$PDF{Page} = $$PDF{RootPage}->new_page;
	$$PDF{Outline} = $$PDF{pdf}->new_outline( Title => 'Outline Page', Destination => $$PDF{Page} );
	my $Line = PDFLinePosition( $PDF, 1 );
	$$PDF{Page}->stringl( $$PDF{Font}, $$PDF{FontSize}, $$PDF{LeftMargin}, $Line, "Page $$PDF{PageNumber}" );
	$$PDF{Page}->stringc( $$PDF{Font}, $$PDF{FontSize}, $$PDF{Center}, $Line, $$PDF{Title} );
	$$PDF{Page}->stringr( $$PDF{Font}, $$PDF{FontSize}, $$PDF{RightMargin}, $Line, "$$PDF{Date}" );
	$$PDF{LineNumber} = 2;
	PDFPrintSubHeader($PDF);
	PDFColumnHeader($PDF);
	$$PDF{LineNumber}++;
	$$PDF{LinePosition} = PDFLinePosition( $PDF, $$PDF{LineNumber} );
}

sub PDFPrintSubHeader {
	my ($PDF) = @_;
	return if !$$PDF{PageSubHeaders};
	foreach my $Header ( @{ $$PDF{PageSubHeaders} } ) {
		$$PDF{LineNumber}++;
		$$PDF{LinePosition} = PDFLinePosition( $PDF, $$PDF{LineNumber} );
		PDFPrint( $PDF, $Header, 'C' );
	}
	$$PDF{LineNumber}++;
}

sub PDFColumnHeader {
	my ($PDF) = @_;
	$$PDF{LineNumber}++;
	$$PDF{LinePosition} = PDFLinePosition( $PDF, $$PDF{LineNumber} );
	PDFPrint( $PDF, $$PDF{ColumnHeader}, '', '', 1 );
	PDFDrawLine( $PDF, $$PDF{LineNumber} + 0.25 );
}

sub PDFDrawLine {
	my ( $PDF, $Line, $Left, $Right ) = @_;
	$Left  = $Left  ? PDFColumnPosition( $PDF, $Left )  : $$PDF{LeftMargin};
	$Right = $Right ? PDFColumnPosition( $PDF, $Right ) : $$PDF{RightMargin};
	$Line = $Line ? PDFLinePosition( $PDF, $Line ) : $$PDF{LinePosition} - 1;
	$$PDF{Page}->line( $Left, $Line, $Right, $Line );
}

sub PDFLinePosition {
	my ( $PDF, $Line ) = @_;
	return $$PDF{PaperLength} - ( $$PDF{FontSize} * $Line ) - $$PDF{Margin};
}

sub PDFColumnPosition {
	my ( $PDF, $Column ) = @_;
	return ( $Column - 1 ) * $$PDF{FontSize} + 1 + $$PDF{Margin};
}

# ---------------------------------------------------------------------------------------------
sub UniqueFilename {
	my ($name,$extension) = @_;
	return $name if $name eq '-';
	my ($dirname,$filename) = (dirname($name),basename($name));
	$filename =~ s"[^a-zA-Z0-9_-]""g;
	$extension = $1 if !$extension and $filename =~ s"\.(...)$"";
	my $filenumber = 0;
	my $file = "$dirname/$filename.$extension";
	while ( -f $file and !unlink $file ) {
		$filenumber++;
		$file = "$dirname/${filename}_$filenumber.$extension";
	}
	return $file;
}

# ---------------------------------------------------------------------------------------------
# End
1;
